function [A, R, B, objects_A, relations, objects_B] = mre(data, no_dims)
%MRE Generic implementation of Matrix Relational Embedding
%
%   [A, R, B, objects_A, relations, objects_B] = mre(data, no_dims)
%
% Performs Matrix Relational Embedding on a collection of given relations.
% The N relations are given in an Nx3 matrix data of which the first column
% indicates the first object A, the second column is the relation R, and
% the third column contains the second object B.
% The 3D matrices A and B contain the learned representations for the two 
% sets of objects, and the 3D matrix R that of the relations. The 
% corresponding objects and relations are returned in objects_A, objects_B, 
% and relations, respectively.
%
%
% (C) Laurens van der Maaten, 2008
% Maastricht University


    % Identify unique objects and relations
    [objects_A, foo, ind1] = unique(data(:,1));
    [relations, foo, ind2] = unique(data(:,2));
    [objects_B, foo, ind3] = unique(data(:,3));
    ind = [ind1 ind2 ind3];
    nA = numel(objects_A);
    nR = numel(relations);
    nB = numel(objects_B);
    
    % Initalize random solution
    A = randn(no_dims, no_dims, nA);
    R = randn(no_dims, no_dims, nR);
    B = randn(no_dims, no_dims, nB);
    
    % Minimization using conjugate gradients
    x = [A(:); R(:); B(:)];
    x = minimize(x, 'mre_grad', -1000, ind, nA, nR, nB, no_dims);
    A = reshape(x(1:nA * no_dims * no_dims), [no_dims no_dims nA]);
    R = reshape(x(1 + nA * no_dims * no_dims:(nA + nR) * no_dims * no_dims), [no_dims no_dims nR]);
    B = reshape(x(1 + (nA + nR) * no_dims * no_dims:end), [no_dims no_dims nB]);
